//********************************************************
//*** STLTutorial.cpp - A concordance application
//*** This application generates a concordance for a given text file. 
//*** A concordance lists each word that occurs in the text and its frequency of occurrence.
//*** The words are sorted in order of decreasing frequency.
//***
//*** by Emil Vassev
//********************************************************

#include "Word.h"

#include <iostream>
#include <fstream>
#include <map>
#include <string>
#include <exception>
#include <vector>
#include <algorithm>
#include <iomanip>

using namespace std;

//global constants
const string EMPTY_WORD = "";
const string EXC_STL_PREFIX = "STL Exception: ";
const string EXC_CPP_PREFIX = "C++ Exception: ";
const string EXC_FILE_OPEN = "Cannot open the file.";
const string EXC_CPP_GENERAL = "Unhandled C++ exception occurred.";
const string MSG_CONC_RESULT = "The concordance for this text file is: ";

//open & close file routines
void openFile(ifstream &ifsInFile)
{
	string sFileName;

	cout << "Please, enter the file name: ";
	cin >> sFileName;

	ifsInFile.open(sFileName.c_str());
}

void closeFile(ifstream &pifsFile)
{
	pifsFile.clear(); 
	pifsFile.close();
}

//overloading stream operator >>
//this version of the operator >> works with the class Word; 
//it discards characters that are not letters and converts all words to lower case before assigning them
ifstream& operator>> (ifstream &pifStream, Word &oWord)
{
    string sTemp;
	char chTemp;
	
	do
	{
			oWord.sWord = EMPTY_WORD;  //initializes the object oWord to "empty" word
			pifStream >> sTemp;        //gets the current word from the stream

			int iLength = (int)sTemp.length();

			for (int i=0; i<iLength; ++i)
			{
				chTemp = tolower(sTemp.at(i)); //converts the current char to lowercase char
				if ((chTemp >= 'a') && (chTemp <= 'z'))
					oWord.sWord.append(1, chTemp); //appends only letters
			}
	}
	while ((oWord.sWord == EMPTY_WORD) && (!pifStream.eof()));//doesn't accept "empty" word

	//ignores the last word if it is "empty" word 
	if ((pifStream.eof()) && (oWord.sWord == EMPTY_WORD))
		pifStream >> oWord.sWord;

	//sets new maxLen if any
	oWord.setMaxWordLen(); 

	return pifStream;
}

//defines a function for sorting pairs by second value in decreasing order
//this function is used by the sort algorithm (see the sort() function below) 
bool decSortOrder(pair<string, int> &pLeft, pair<string, int> &pRight) 
{
	return (pLeft.second > pRight.second);
}

int main()
{
	try
	{
		ifstream ifsInFile; 
		Word currWord; //we use one sinle Word object for all the words
		map<string, int> mCounters;
		vector<pair<string, int> > vWords;

		openFile(ifsInFile);

		//proceed only if the ifstream is open (not fail) 
		if (ifsInFile.fail()) 
			throw exception(EXC_FILE_OPEN.c_str());

		while (ifsInFile >> currWord)  //filters the words from non-letter characters 
			++mCounters[currWord.getWord()];

		closeFile(ifsInFile);

		//elements in mCounters map are already sorted by key and sort() will not work on mCounters map;
		//hence, we copy mCounters map into vWords vector, whcih could be sorted;
		//we use back_inserter() function to allocate memory for the vector vWords;
		//this function causes the containers push_back operation to be invoked;
		copy(mCounters.begin(), mCounters.end(), back_inserter(vWords)); 
		
		mCounters.clear(); 
		
		//uses decSortOrder to sort the vector vWords by second pairs' value 
		sort(vWords.begin(), vWords.end(), decSortOrder);    

		//prints out the concordance result
		cout << endl << MSG_CONC_RESULT << endl << endl;
		for ( vector<pair<string, int> >::const_iterator it = vWords.begin(); it != vWords.end(); ++it)
		{
			//for formatting the output, we use the max length (the length of the longest word)
			cout << left << setw(static_cast<streamsize>(currWord.getMaxWordLen())) << it->first << "\t" << it->second << endl;
		}

		vWords.clear();

		return 0;
	}
	catch (exception& ex) // Handle STL exceptions
	{
		cerr << EXC_STL_PREFIX << ex.what() << endl;
		return -1;
	}
	catch (...) // Handle all C++ exceptions
	{
		cerr << EXC_CPP_PREFIX << EXC_CPP_GENERAL << endl;
		return -1;
	}
}